import os
import sys

import pyhectiqlab
from pyhectiqlab.metrics import MetricsManager
from pyhectiqlab.events_manager import EventsManager
from pyhectiqlab.stream import LogHandler
from pyhectiqlab.buffer import Buffer
from pyhectiqlab.watermark import Watermark
from pyhectiqlab.artifacts import SharedArtifactsManager
from pyhectiqlab.notebooks import get_notebook_file
from pyhectiqlab.utils import list_all_files_in_dir

from pyhectiqlab.mlmodels import download_mlmodel as ops_download_mlmodel
from pyhectiqlab.mlmodels import upload_mlmodel as ops_upload_mlmodel
from pyhectiqlab.datasets import download_dataset as ops_download_dataset
from pyhectiqlab.datasets import upload_dataset as ops_upload_dataset


from functools import partial
from sys import getsizeof
import packaging 
import tempfile
import logging
import shutil

class Run():
    
    def __init__(self, name:str, project:str=None, project_name:str=None, should_create_project = False, dry=False):
        """
        dry: bool: If true, the run does not send anything.
        should_create_project: bool. If true, a project is created if if does not exists
        """
        self._mock = dry
        self.events_manager = EventsManager(self._mock)
        # Check the python version
        self.events_manager.compare_python_version()    
        if project_name is not None:
            print('Argument `project_name` will be remove in a future release. Please use `project` instead.')
            project = project_name
        if dry:
            print("Dry mode activated.")
            self.run_view = {"id": "null"}
            self._mock = True
            self.events_manager = EventsManager(self._mock)
        elif self.events_manager.is_logged()==False:
            print("User not authentificated. Use `pyhectiqlab.login()` for a web prompt or `hectiqlab login` in command line.")
            self._failed_init()
        else:
            # try:
            if project is None:
                projects = self.events_manager.add_event("get_all_projects", args=(), auth=True, async_method=False)
                if len(projects["result"])==1:
                    project = projects["result"][0]["name"]
                    print(f"Connecting to project {project}.")
                    self.project_view = projects["result"][0]
                else:
                    print("User has access to multiple projects. Switching to dry mode. \n--------\nFor active mode, please use a project:")
                    for el in projects["result"]:
                        print(f"{el['name']}")
                    self._failed_init()
                    pass
            else:
                self.project_view = self.events_manager.add_event("create_project", args=(project, should_create_project), auth=True, async_method=False)
                if self.project_view is None:
                    print("Project does not exists. Dry mode activated. \n--------\nUse should_create_project=True to create a project.")
                    self._failed_init()
                    pass
                if self.project_view['status_code']==401:
                    self._failed_init()
                    pass
            if project:
                self.run_view = self.events_manager.add_event("create_run", args=(name, project), auth=True, async_method=False)
                try:
                    if self.run_view["is_new"]==False:
                        print(f"Attaching to an existing run (ID :{self.run_view['p_id']}).")
                    if self.run_view['readonly']:
                        print(f"Run is read only.")
                    self._read_only = self.run_view['readonly']
                except:
                    print("An error occured while creating the run. Switching to dry mode.")
                    self._failed_init()

                if "id" not in self.run_view:
                    print("An error occured while creating the run. Switching to dry mode.")
                    self._failed_init()

        self.name = name
        self.metrics_manager = MetricsManager(run_id=self.run_view["id"], push_method=self._push_metrics)
        self.metrics_manager.set_aggr('mean')
        self.metrics_manager.update_cache_settings(max_cache_timeout=20,
                                                    min_cache_flush_delay=5, 
                                                    max_cache_length=25)
        self.log_recorder = None
        self.is_recording_logs = False
        self.logs_buffer = Buffer(push_method=self._push_logs)
        return

    @property
    def read_only(self):
        return self._read_only

    def set_metrics_aggregate_method(self, aggr):
        """
        Set the aggregate method for the metrics manager. 

        aggr [str]: One of ['none', 'sum', 'max', 'mean']
        """
        print('`set_metrics_aggregate_method` will be remove in a future release. Please use `set_metrics_cache_settings` instead.')
        self.metrics_manager.set_aggr(aggr)

    def set_metrics_cache_settings(self, aggr: str = 'mean', max_cache_timeout=None, max_cache_length=None, min_cache_flush_delay=None):
        """
        Update the settings of the metrics manager.

        max_cache_timeout [int]: Time interval for flushing the cache (seconds)  (default is 20)
        max_cache_length [int]: Max number of elements before flushing the cache (default is 25)
        min_cache_flush_delay [float]: Minimum delay between to successive data send (for the same key). 
                If the number of matrix is added faster than this delay, the cache will increase to
                respect the delay. (default is 5 seconds)
        aggr [str]: The aggregate method. One of ['none', 'sum', 'max', 'mean']  (default is mean)
        """
        self.metrics_manager.set_aggr(aggr)
        self.metrics_manager.update_cache_settings(max_cache_timeout=max_cache_timeout, 
                                                    max_cache_length=max_cache_length,
                                                    min_cache_flush_delay=min_cache_flush_delay)

    def _failed_init(self):
        self.run_view = {"id": "null"}
        self._mock = True
        self.events_manager = EventsManager(self._mock)
        self._read_only = True

    def add_artifacts(self, filepath: str, step: int = None):
        """Log a file as an artifacts. If a file already exists with the name, it will be overwritten.

        filepath [str] : Path to the file.
        """
        assert self._read_only==False, 'Run is in read only.'
        filename = os.path.basename(filepath)
        num_bytes = os.path.getsize(filepath)
        content_bytes = open(filepath, "rb")
        args = (self._id, filename, content_bytes, num_bytes, step)
        self.events_manager.add_event("add_file", args)

    def add_mlmodel(self, source_path: str, name: str, version: str = None, short_description: str = None,
                push_dir: bool = False, show_filenames: bool = False):
        """Add a mlmodel
        """
        assert self._read_only==False, 'Run is in read only.'
        ops_upload_mlmodel(source_path=source_path, 
            mlmodel_name=name,
             run_id=self._id, 
             version=version,
             short_description=short_description,
             push_dir=push_dir)

    def download_mlmodel(self, mlmodel_name: str, version: str = None, save_path: str = './', overwrite:bool = False):

        dir_path = ops_download_mlmodel(mlmodel_name=mlmodel_name, 
                        project_id=self.project_view["id"], 
                        version=version, 
                        save_path=save_path, 
                        overwrite=overwrite)
        if dir_path is not None:
            print(dirpath) 

    def add_dataset(self, source_path: str, name: str, version: str = None, short_description: str = None,
                push_dir: bool = False, show_filenames: bool = False):
        """Add a mlmodel
        """
        assert self._read_only==False, 'Run is in read only.'
        ops_upload_dataset(source_path=source_path, 
            dataset_name=name,
             run_id=self._id, 
             version=version,
             short_description=short_description,
             push_dir=push_dir)

    def download_dataset(self, dataset_name: str, version: str = None, save_path: str = './', overwrite:bool = False):

        dir_path = ops_download_dataset(dataset_name=dataset_name, 
                        project_id=self.project_view["id"], 
                        version=version, 
                        save_path=save_path, 
                        overwrite=overwrite)
        if dir_path is not None:
            print(dirpath)        

    def add_config(self, config, prefix=""):
        assert self._read_only==False, 'Run is in read only.'
        config.push_as_meta(self, prefix=os.path.join("config/",prefix))

    def add_tf_model(self, model, filename, step: int = None):
        assert self._read_only==False, 'Run is in read only.'
        save_path = tempfile.mkdtemp()
        model.save_weights(f"{save_path}/{filename}")
        p = tempfile.mkdtemp()
        shutil.make_archive(f"{p}/{filename}", 'zip', save_path)
        self.add_artifacts(f"{p}/{filename}.zip", step=step)

    def add_directory_as_zip_artifact(self, dirpath, step: int = None):
        assert self._read_only==False, 'Run is in read only.'
        p = tempfile.mkdtemp()
        filename = os.path.basename(dirpath)
        shutil.make_archive(f"{p}/{filename}", 'zip', root_dir=dirpath)
        self.add_artifacts(f"{p}/{filename}.zip", step=step)

    def add_shared_artifact(self, filename):
        """Push an artifact to be shared accross the project.
        """
        assert self._read_only==False, 'Run is in read only.'
        m = SharedArtifactsManager()
        m.push_artifact(filepath=filename, project_id=self.project_view["id"])

    def list_shared_artifact(self):
        """Returns a list of the shared artifacts for this project.
        """
        m = SharedArtifactsManager()
        return m.list_artifacts(project_id=self.project_view["id"])

    def add_metrics(self, key, value, step):
        """Add a metrics. The metrics is appended even if there already exists a value
        for the same key and step. 

        key [str]: Key used for the metrics. The format with backlash such as `metrics/train`
            is recognized to organize the metrics.

        value [float or string]. If tensor, is casted to float.
        step [int]: Training step. Casted to integer.
        """
        assert self._read_only==False, 'Run is in read only.'
        self.metrics_manager.add(key, float(value), int(step))

    def add_tag(self, name: str, description: str = None, color: str = None):
        """Add tag to a run. If the tag does not already exists by this name,
        a new tag is created with the given description and color.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, name, description, color)
        self.events_manager.add_event("add_tag", args)

    def add_package_versions(self, packages=None, with_sys=True, with_python=True, with_git=True):
        """Save the version of the imported packages.

        Use:
        add_package_versions(globals())
        """
        assert self._read_only==False, 'Run is in read only.'
        water = Watermark()
        versions = {}
        if with_python:
            versions["python"] = water.get_pyversions()
        if with_sys:
            versions["system"] = water.get_sysinfo()
        if packages:
            versions["packages"] = water.get_all_import_versions(packages)
        if with_git:
            git_info = water.get_git_info()
            if git_info:
                versions["git"] = water.get_git_info()

        args = (self._id, versions)
        self.events_manager.add_event("push_package_versions", args)

    def add_package_repo_state(self, package: str = None):
        """Save the branch/commit/origin of the git repo in which
        a package is located.

        Example:
            self.add_package_repo_state('pyhectiqlab')
        """
        assert self._read_only==False, 'Run is in read only.'
        water = Watermark()
        result = water.get_repo_info(package)
        data = {package: result}
        args = (self._id, data)
        self.events_manager.add_event("push_git_package_state", args)


    def add_current_notebook(self, stamp: str = 'datetime'):
        """If the run is executed in a jupyter notebook, this
        will quickly add your current notebook as an artifact.

        stamp [str]: Add a stamp on the notebook name.
                    One of 'datetime', 'date', None
        """
        assert self._read_only==False, 'Run is in read only.'
        filename = get_notebook_file(stamp)
        self.add_artifacts(filename)

    def failed(self):
        """Set the run to the failed status.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, "failed")
        self.events_manager.add_event("set_run_status", args)

    def _stopped(self):
        """Set the run to the stopped status.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, "stopped")
        self.events_manager.add_event("set_run_status", args)

    def completed(self):
        """Set the run to the completed with success status.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, "completed")
        self.events_manager.add_event("set_run_status", args)

    def pending(self):
        """Set the run to the pending status.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, "pending")
        self.events_manager.add_event("set_run_status", args)

    def running(self):
        """Set the run to the running status.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, "running")
        self.events_manager.add_event("set_run_status", args)

    def training(self):
        """Set the run to the training status.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, "training")
        self.events_manager.add_event("set_run_status", args)

    def add_meta(self, key, value):
        """Add or update a meta information in the format of
        key/value.
        """ 
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, key, value)
        self.events_manager.add_event("push_meta", args)

    def set_note(self, text):
        """Add or update a custom note field for the run.
        """
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, text)
        self.events_manager.add_event("set_note", args)

    def start_recording_logs(self, logger_name=None):
        """Start tailing the logs on a logger name.
        If logger_name is None, the default __name__ will be taken and
        you may use the returned logger.
        """
        assert self._read_only==False, 'Run is in read only.'
        self.is_recording_logs = True
        self.logs_buffer.start(key="logs")
        
        print("Start recording logs.")
        if (self.log_recorder is None) or self.logger_name!=logger_name:
            func = partial(self.logs_buffer.add, key="logs")
            self.log_recorder = LogHandler(func)
            console  = logging.StreamHandler()  
            console.setLevel(logging.DEBUG)

            name = logger_name
            if name is None:
                name = __name__
            self.logger_name = name

            logger = logging.getLogger(name)
            logger.setLevel(logging.DEBUG)
            logger.addHandler(self.log_recorder)

            # Handle all exceptions
            def handle_exception(exc_type, exc_value, exc_traceback):
                if issubclass(exc_type, KeyboardInterrupt):
                    self._stopped()
                else:
                    self.add_tag(name="failed")
                    self.failed()
                    
                logger.error("Exception", exc_info=(exc_type, exc_value, exc_traceback))
                # Flush cache before leaving
                self.logs_buffer.flush_cache()
                self.metrics_manager.flush_cache()
                sys.exit(0)                
                
            sys.excepthook = handle_exception

            return logger
        else:
            return logging.getLogger(self.logger_name)

    def stop_recording_logs(self):
        """Stop recording the logs from the logger. 
        """
        assert self._read_only==False, 'Run is in read only.'
        self.is_recording_logs = False
        self.logs_buffer.stop(key="logs")

    def _push_logs(self, key, msg):
        assert self._read_only==False, 'Run is in read only.'
        if key=="logs":
            if self.is_recording_logs:
                args = (self._id, "".join(msg))
                self.events_manager.add_event("append_logs", args)

    def _push_metrics(self, key, values):
        assert self._read_only==False, 'Run is in read only.'
        args = (self._id, key, values)
        self.events_manager.add_event("push_metrics", args)

    @property
    def _id(self):
        return self.run_view["id"]
    


import platform
import types
from git import Repo, InvalidGitRepositoryError
from multiprocessing import cpu_count
import subprocess
import importlib
import inspect
import socket

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata

"""
Source: https://github.com/rasbt/watermark

New BSD License

Copyright (c) 2014-2018, Sebastian Raschka.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of watermark nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

class Watermark():

    def get_packages(self, pkgs):
        packages = pkgs.split(",")
        return {package: self.get_package_version(package)
                for package in packages}

    def get_package_version(self, pkg_name):
        """Return the version of a given package"""
        if pkg_name == "scikit-learn":
            pkg_name = "sklearn"
        try:
            imported = importlib.import_module(pkg_name)
        except ImportError:
            version = "not installed"
        else:
            try:
                version = importlib_metadata.version(pkg_name)
            except importlib_metadata.PackageNotFoundError:
                try:
                    version = imported.__version__
                except AttributeError:
                    try:
                        version = imported.version
                    except AttributeError:
                        try:
                            version = imported.version_info
                        except AttributeError:
                            version = "unknown"
        return version

    def get_pyversions(self):
        return {
            "Python implementation": platform.python_implementation(),
            "Python version": platform.python_version()
        }

    def get_sysinfo(self):
        return {
            "Compiler": platform.python_compiler(),
            "OS": platform.system(),
            "Release": platform.release(),
            "Machine": platform.machine(),
            "Processor": platform.processor(),
            "CPU cores": cpu_count(),
            "Architecture": platform.architecture()[0],
            "Hostname": socket.gethostname(),
        }

    def get_git_info(self):
        data = {"hash": self.get_commit_hash(), 
        "origin": self.get_git_remote_origin(), 
        "branch": self.get_git_branch()}
        if data["hash"] != "":
            return data
        return None

    def get_commit_hash(self):
        process = subprocess.Popen(
            ["git", "rev-parse", "HEAD"], shell=False, stdout=subprocess.PIPE
        )
        git_head_hash = process.communicate()[0].strip()
        return git_head_hash.decode("utf-8")

    def get_git_remote_origin(self):
        process = subprocess.Popen(
            ["git", "config", "--get", "remote.origin.url"],
            shell=False,
            stdout=subprocess.PIPE,
        )
        git_remote_origin = process.communicate()[0].strip()
        return git_remote_origin.decode("utf-8")

    def get_git_branch(self):
        process = subprocess.Popen(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            shell=False,
            stdout=subprocess.PIPE,
        )
        git_branch = process.communicate()[0].strip()
        return git_branch.decode("utf-8")

    def get_repo(self, package_name):
        spec = importlib.util.find_spec(package_name)
        if spec is None:
            return
            
        path = spec.origin
        try:
            repo = Repo(path, search_parent_directories=True)
        except InvalidGitRepositoryError:
            return
        
        return repo
    
    def get_repo_info(self, package_name):
        repo = self.get_repo(package_name)
        assert repo is not None, 'Unable to find the repo.'
        
        return {
            'branch': repo.active_branch.name,
            'commit': repo.active_branch.commit.hexsha,
            'remote': repo.remotes.origin.url
        }
        
    def get_all_import_versions(self, global_namespaces):

        to_print = {}
        imported_pkgs = {
            val.__name__.split(".")[0]
            for _, val in global_namespaces.items()
            if isinstance(val, types.ModuleType)
        }
        submodules = {
            inspect.getmodule(val).__name__.split(".")[0]
            for _, val in global_namespaces.items()
            if (inspect.getmodule(val) is not None)
        }
        imported_pkgs = set.union(submodules, imported_pkgs)

        imported_pkgs.discard("builtins")
        imported_pkgs.discard("IPython")
        imported_pkgs.discard("__main__")
        
        for pkg_name in imported_pkgs:
            pkg_version = self.get_package_version(pkg_name)
            if pkg_version not in ("not installed", "unknown"):
                to_print[pkg_name] = pkg_version
        return to_print


import os
import requests
import json
import pyhectiqlab.settings as settings
import threading
import logging
import time
from typing import List

logger = logging.getLogger('hectiqlab')

def lab_request(path, request_type, form_data=None, query_args=None, body_args=None, token=None):

    url = settings.server_url + path
    if query_args:
        url += "?"
        for key, val in query_args.items():
            if val is not None:
                url += f"{key}={val}&"

    headers = {}
    if token:
        headers = {"Authorization": f"Bearer {token}"}
    body = None
    if body_args:
        body = json.dumps(body_args)
    files = None
    if form_data:
        body = form_data
        headers["content-type"] = 'application/x-www-form-urlencoded'
    if request_type=="GET":
        call = requests.get
    elif request_type=="PUT":
        call = requests.put
    elif request_type=="POST":
        call = requests.post
    try:
        res = call(url, data=body, headers=headers, files=files)
        logger.debug('LABREQUEST:')
        logger.debug(url)
        logger.debug(res.status_code)
        if res.status_code==401:
            logger.debug(res.json()["detail"])
            print(res.json()["detail"])
            return {"status_code": 401, "detail": res.json()["detail"]}
        else:
            data = res.json()
            logger.debug(res.json())
            if isinstance(data, list):
                data = {"result": data}
            data["status_code"] = res.status_code
            return data
    except:
        return {"status_code": 400}

def refresh_token(refresh_token):
    res = lab_request(path="/update-token", 
                request_type="POST", 
                body_args=dict(refresh_token=refresh_token))
    return res

def login(username, password):
    res = lab_request(path="/login", 
                request_type="POST", 
                form_data=dict(username=username, password=password))
    return res

def fetch_minimum_python_version():
    res = lab_request(path="/minimum-supported-python-version", 
                request_type="GET")
    return res

def create_project(name, force_create, token):
    res = lab_request(path="/projects/", 
                request_type="POST", 
                token=token,
                body_args=dict(name=name, force_create=force_create))
    return res

def list_shared_artifacts(project_id, token):
    res = lab_request(path="/artifacts/shared", 
                request_type="GET", 
                token=token,
                query_args=dict(project_id=project_id))
    return res

def get_artifact_signed_url(artifact_uuid, token):
    res = lab_request(path=f"/artifacts/url", 
                request_type="GET", 
                token=token,
                query_args=dict(artifact_uuid=artifact_uuid))
    return res

def create_run(name, project_name, token):
    res = lab_request(path="/runs/", 
                request_type="POST", 
                token=token,
                body_args=dict(name=name, project_name=project_name))
    return res

def get_existing_run_info(run_id, token):
    res = lab_request(path=f"/runs/{run_id}", 
                request_type="GET", 
                token=token)
    return res

def push_metrics(run_id, metrics_name, values, token):
    res = lab_request(path=f"/metrics/{run_id}/add",
                     request_type="POST",
                     token=token,
                     body_args=dict(name=metrics_name, values=values))
    return res

def set_run_status(run_id, status, token):
    res = lab_request(path=f"/runs/{run_id}/status",
                     request_type="POST",
                     token=token,
                     body_args=dict(status=status))
    return res

def push_meta(run_id, key, value, token):
    res = lab_request(path=f"/runs/{run_id}/meta",
                     request_type="POST",
                     token=token,
                     body_args=dict(meta={key:value}))
    return res

def push_package_versions(run_id, data, token):
    res = lab_request(path=f"/runs/{run_id}/packages",
                     request_type="POST",
                     token=token,
                     body_args=dict(packages=data))
    return res

def push_git_package_state(run_id, data, token):
    res = lab_request(path=f"/runs/{run_id}/git_package_meta",
                     request_type="POST",
                     token=token,
                     body_args=dict(gitmeta=data))
    return res

def set_note(run_id, note, token):
    res = lab_request(path=f"/runs/{run_id}/comment",
                     request_type="POST",
                     token=token,
                     body_args=dict(text=note))
    return res

def add_tag(run_id, name, description, color, token):
    res = lab_request(path=f"/runs/{run_id}/tags",
                     request_type="POST",
                     token=token,
                     body_args=dict(tag_name=name, tag_description=description, tag_color=color, can_remove=False))
    return res

def append_logs(run_id, logs, token):
    if logs is None:
        return
    res = lab_request(path=f"/runs/{run_id}/logs",
                     request_type="POST",
                     token=token,
                     body_args=dict(logs=logs))
    return res

def add_file(run_id, filename: str, content_bytes: bytes, num_bytes: int = None, step: int = None, token: str = None):

    # Get policy
    args = dict(run_id=run_id, filename=filename)
    if step is not None:
        args['step'] = step
    if num_bytes:
        args["num_bytes"] = num_bytes
    res = lab_request(path=f"/runs/{run_id}/file_policy",
                     request_type="GET",
                     token=token,
                     query_args=args)
    try:
        policy, bucket_name, route = res["policy"], res["bucket_name"], res['route']
        # Send file
        files = {"file": (bucket_name, content_bytes)}
        requests.post(policy["url"], data=policy["fields"], files=files)

        # Callback
        res = lab_request(path=route,
                     request_type="GET",
                     token=token)
        return res
    except:
        # Unable
        return res

def get_mlmodel_download_info(mlmodel_name, project_id, version, token):
    res = lab_request(path=f"/mlmodels/download_info/",
                     request_type="GET",
                     token=token,
                     query_args=dict(mlmodel_name=mlmodel_name, project_id=project_id, version=version))
    return res

def get_dataset_download_info(dataset_name, project_id, version, token):
    res = lab_request(path=f"/datasets/download_info/",
                     request_type="GET",
                     token=token,
                     query_args=dict(dataset_name=dataset_name, project_id=project_id, version=version))
    return res

def add_mlmodel(run_id, filenames: List[str], full_paths: List[str], num_bytes: List[int], name: str, short_description: str, version: str, token: str = None):

    # Get policy
    body = dict(run_id=run_id, name=name, filenames=filenames, 
        short_description=short_description, 
        num_bytes=num_bytes, 
        version=version)

    res = lab_request(path=f"/runs/{run_id}/mlmodel",
                     request_type="POST",
                     token=token,
                     body_args=body)
    if res.get('status_code') == 400:
        print(res.get('detail'))
        return

    for filename, content in res['policies'].items():
        print(f'Uploading {filename}')
        if filename is None:
            continue
        content_bytes = open(full_paths[filename], 'rb')

        logger.debug(filename, content)
        policy, bucket_name, route = content["policy"], content["bucket_name"], content['route']

        # Send file
        file = {"file": (bucket_name, content_bytes)}

        logger.debug(policy["url"], policy["fields"])
        requests.post(policy["url"], data=policy["fields"], files=file)

        # Callback
        lab_request(path=route,
                     request_type="GET",
                     token=token)
    return res

def add_dataset(run_id, filenames: List[str], full_paths: List[str], num_bytes: List[int], name: str, short_description: str, version: str, token: str = None):

    # Get policy
    body = dict(run_id=run_id, name=name, filenames=filenames, 
        short_description=short_description, 
        num_bytes=num_bytes, 
        version=version)

    res = lab_request(path=f"/runs/{run_id}/dataset",
                     request_type="POST",
                     token=token,
                     body_args=body)
    if res.get('status_code') == 400:
        print(res.get('detail'))
        return

    for filename, content in res['policies'].items():
        print(f'Uploading {filename}')
        if filename is None:
            continue
        content_bytes = open(full_paths[filename], 'rb')

        logger.debug(filename, content)
        policy, bucket_name, route = content["policy"], content["bucket_name"], content['route']

        # Send file
        file = {"file": (bucket_name, content_bytes)}

        logger.debug(policy["url"], policy["fields"])
        requests.post(policy["url"], data=policy["fields"], files=file)

        # Callback
        lab_request(path=route,
                     request_type="GET",
                     token=token)
    return res


def add_shared_artifact(project_id: str, filename: str, content_bytes: bytes, num_bytes: int = None, token: str = None):

    # Get policy
    args = dict(project_id=project_id, filename=filename)
    if num_bytes:
        args["num_bytes"] = num_bytes
    res = lab_request(path=f"/artifacts/shared",
                     request_type="POST",
                     token=token,
                     query_args=args)

    try:
        policy, bucket_name, route = res["policy"], res["bucket_name"], res['route']
        # Send file
        files = {"file": (bucket_name,content_bytes)}
        requests.post(policy["url"], data=policy["fields"], files=files)

        # Callback
        res = lab_request(path=route,
                     request_type="GET",
                     token=token)
        return res
    except:
        # Unable
        return res

def get_shared_artifacts_download_link(artifact_uuid, token):
    res = lab_request(path=f"/artifacts/url",
                     request_type="GET",
                     token=token,
                     query_args=dict(artifact_uuid=artifact_uuid))
    return res

def get_all_projects(token):
    res = lab_request(path=f"/projects/all",
                     request_type="GET",
                     token=token)
    return res


import os
from pathlib import Path
import logging
import ast
import requests
from tqdm import tqdm

from pyhectiqlab.config import Config
from pyhectiqlab.events_manager import EventsManager

logger = logging.getLogger('pyhectiqlab')
logger.setLevel(logging.WARNING)

def load_event_manager():
    events_manager = EventsManager()
    assert events_manager.is_logged(), "User not authentificated"
    return events_manager

def get_single_user_project():

    manager = load_event_manager()
    projects = manager.add_event("get_all_projects", args=(), auth=True, async_method=False)
    if len(projects["result"])==1:
        project = projects["result"][0]
        logger.info(f"Connecting to project {project['name']}.")
        return project
    else:
        logger.error("User has access to multiple projects.")
        return

def run_config(run_id: str):
    """Fetch the config file of an existing run.
    """
    events_manager = load_event_manager()
    run_view = events_manager.add_event('get_existing_run_info', 
        (run_id, ), 
        auth=True, async_method=False)
    logger.info(f"Run {run_id}: {run_view['name']}")
    if 'custom_fields' not in run_view:
        return Config()

    return convert_custom_fields_to_config(run_view['custom_fields'])

def download_existing_run_artifact(artifact_uuid: str, savepath: str = "./"):
    events_manager = load_event_manager()
    data = events_manager.add_event('get_artifact_signed_url', 
                                    (artifact_uuid, ), 
                                    auth=True, async_method=False)
    if 'detail' in data:
        logger.error(data['detail'])
        return

    if 'url' not in data:
        logger.error('The answer is unexpected.')
        return

    url = data['url']
    artifact_name = data['name']
    response = requests.get(url, stream=True)
    total_size_in_bytes= int(response.headers.get('content-length', 0))
    block_size = 1024
    progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
    
    path = os.path.join(savepath, artifact_name)
    with open(path, 'wb') as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        logger.error("ERROR, something went wrong when downloading the file.")
    
    logger.info(f'Content saved at {path}')
    return path

def convert_custom_fields_to_config(data: dict):
    """Convert a dict representation of a config file
    into a Config object.

    data [dict]: Dict representation of a config

    Returns:
        config [Config]
    """
    config = Config()
    
    if data is None:
        return config
    
    c = config
    for key in data:
        keys = key.split("/")
        for i,k in enumerate(keys):
            if i==0:
                continue
            if i<len(keys)-1:
                if k not in c.dict:
                    setattr(c, k, Config())
                c = getattr(c, k)
            else:
                d = data[key]
                try:
                    d = ast.literal_eval(d)
                except:
                    True
                setattr(c, k, d)
        c = config
    return config

def list_all_files_in_dir(local_folder: str):
    path_object = Path(local_folder)
    filenames = []
    for el in os.walk(local_folder):
        for f in os.listdir(el[0]):
            complete_path = os.path.join(el[0], f)
            if os.path.isdir(complete_path)==False:
                if os.path.isfile(complete_path):
                    filenames.append(complete_path)
    return filenames


import os
import json
import jwt
import datetime as dt
import pyhectiqlab.ops as ops

from http.server import BaseHTTPRequestHandler, HTTPServer
from webbrowser import open_new
from pathlib import Path
import urllib.parse as urlparse
from urllib.parse import parse_qs

import pyhectiqlab.settings as settings

class HTTPServerHandler(BaseHTTPRequestHandler):
    """
    HTTP Server callbacks to handle OAuth redirects
    """
    def __init__(self, request, address, server):
        super().__init__(request, address, server)

    def do_GET(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        if ("access_token" in self.path):
            parsed = urlparse.urlparse(self.path)
            s = parse_qs(parsed.query)
            self.server.access_token = s["access_token"]
            self.server.refresh_token = s["refresh_token"]
            self.server.expiration = s["expiration"]
            self.wfile.write(bytes('<html><h1>You may now close this window.'+ '</h1></html>', 'utf-8'))
        else:
            self.wfile.write(bytes('<html><h1>Hum. Cannot find the token...'+ '</h1></html>', 'utf-8'))

    def log_message(self, format, *args):
        # Mute
        return

class AuthProvider:
    """
    Class used to handle oAuth
    """
    def __init__(self, port=8082):
        self.port = str(port)
        self._load_tokens()

        # Refresh token if access_token is expired
        if self.tokens:
            if "access_token" in self.tokens:
                if self._is_expired(self.tokens["access_token"])==True:
                    self._refresh_token()

        return
    
    def is_logged(self):
        if self.tokens is None:
            return False
        
        if "access_token" in self.tokens:
            if self._is_expired(self.tokens["access_token"])==False:
                return True
            else:
                self._refresh_token()
                return self.is_logged()
        return False
    
    def login(self):
        if self.is_logged():
            return 
        self._refresh_token()
    
    def logout(self):
        self.tokens = None
        self._save_local_tokens({})

    def login_with_password(self, username:str, password:str):
        tokens = ops.login(username, password)
        if tokens.get('status_code')==401:
            return False
        self._save_local_tokens(tokens)
        self.tokens = tokens
        return True

    @property
    def token(self):
        self.login()
        return self.tokens["access_token"]
    
    def _get_expiration_date(self, token):
        if token is None:
            return None
            
        return jwt.decode(token, verify=False)['exp']
    
    def _is_expired(self, token):
        d = self._get_expiration_date(token)
        if d is None:
            return True
        return dt.datetime.now().timestamp() > d
    
    def _get_access_token(self):
        """
        Get access token from server
        """
        httpServer = HTTPServer(('localhost', int(self.port)),
                lambda request, address, server: HTTPServerHandler(request, address, server))
       
        access_uri = f"{settings.app_url}/api-login?redirectURL=http://0.0.0.0:{self.port}/"
        open_new(access_uri)
        # This function will block until it receives a request
        httpServer.handle_request()
        tokens = {"access_token": httpServer.access_token[0], 
         "refresh_token": httpServer.refresh_token[0], 
         "expiration" :httpServer.expiration[0]}
    
        self._save_local_tokens(tokens)
        self.tokens = tokens
        
        # Close socket (server will be shutdown at return)
        httpServer.socket.close()
        print("Login completed.")
        return 

    def _refresh_token(self):
        if self.tokens is None or self.tokens=={}:
            self._get_access_token()
            return

        if "refresh_token" in self.tokens:
            tokens = ops.refresh_token(self.tokens["refresh_token"])
            self._save_local_tokens(tokens)
            self.tokens = tokens
            return
        
        return

    def _load_tokens(self):
        self.tokens = None
        if os.path.isfile(self.tokens_path):
            file = open(self.tokens_path, "rb")
            self.tokens = json.load(file)
            file.close()
        return

    def _save_local_tokens(self, tokens):
        """
        Save the tokens locally
        """
        if not os.path.exists(self.tokens_folder):
            os.mkdir(self.tokens_folder)

        f = open(self.tokens_path, 'w')
        f.write(json.dumps(tokens))
        f.close()

    @property
    def tokens_path(self):
        home = str(Path.home())
        config_folder = os.path.join(home, '.hectiqlab')
        return os.path.join(config_folder, 'creds.json')
    
    @property
    def tokens_folder(self):
        home = str(Path.home())
        config_folder = os.path.join(home, '.hectiqlab')
        return config_folder



import threading

import pyhectiqlab
import pyhectiqlab.ops as ops
from pyhectiqlab.auth import AuthProvider

from packaging import version
import logging
logger = logging.getLogger('pyhectiqlab')

class EventsManager():
    def __init__(self, mock=False):
        self.auth_provider = AuthProvider()
        self.mock = mock
        events = []

    def is_logged(self):
        return self.auth_provider.is_logged()

    def refresh_token(self):
        self.auth_provider._refresh_token()
        return

    def compare_python_version(self, on_failed = None):
        """Compare the current python version and the
        required python version.
        """
        def threading_method():
            res = ops.fetch_minimum_python_version()
            required_version = res['version']
            if version.parse(pyhectiqlab.__version__) < version.parse(required_version):
                logger.error(f'Your pyhectiqlab version ({pyhectiqlab.__version__}) is obsolete (required: {required_version}).')
                logger.error('Update with:')
                logger.error('`pip install --upgrade pyhectiqlab`')
                logger.error('Some features may break until you upgrade.')
                if on_failed is not None:
                    on_failed()
            return res

        x = threading.Thread(target=threading_method)
        x.start()
        return 

    def add_event(self, task_name, args, auth=True, async_method=True, repeat_if_401=True):
        if self.mock :
            return
        method = getattr(ops, task_name)

        def threading_method(*args, **kwargs):
            res = method(*args, **kwargs)
            if res["status_code"]==401:
                if repeat_if_401:
                    self.auth_provider._refresh_token()
                    self.add_event(task_name, args, auth=auth, async_method=False, repeat_if_401=False)
            return res

        kwargs = {}
        if auth:
            kwargs = {"token": self.auth_provider.token}

        if async_method:
            x = threading.Thread(target=threading_method, args=args, kwargs=kwargs)
            x.start()
            return
        else:
            return method(*args, **kwargs)




from collections import defaultdict
from pyhectiqlab.timer import RepeatedTimer

import numpy as np
import threading
import pyhectiqlab.ops as ops
import time

class MetricsManager():

    def __init__(self, run_id,
        max_cache_timeout=5,  # seconds
        max_cache_length=100, # Number of elements
        min_cache_flush_delay=5, # Minimum number of seconds between a cache flush
        aggregate_metrics="none", # Wheter to push all data in cache or aggregate
        push_method=None):
        self._run = run_id
        self.cache = defaultdict(list)
        self.cache_last_push = {}

        self.max_cache_length = max_cache_length
        self.min_cache_flush_delay  = min_cache_flush_delay
        self.push_method = push_method
        self.set_aggr(aggregate_metrics)

        self.stop_flag = threading.Event()
        self.timer = RepeatedTimer(self.stop_flag, self.flush_cache, max_cache_timeout)
        self.timer.start()
        return

    def update_cache_settings(self, max_cache_timeout, max_cache_length, min_cache_flush_delay):
        if max_cache_timeout is not None:
            self.max_cache_timeout = max_cache_timeout
        if max_cache_length is not None:
            self.max_cache_length = max_cache_length
        if min_cache_flush_delay is not None:
            self.min_cache_flush_delay = min_cache_flush_delay
            
    def set_aggr(self, new_value):
        vals = ['none', 'sum', 'max', 'mean']
        assert new_value in vals, f"Aggr must be in {vals}"
        self.aggregate_metrics = new_value

    def add(self, key, value, step):
        if value is None:
            return
        self.cache[key].append((float(step), float(value)))
        if len(self.cache[key])>self.max_cache_length:
            if key in self.cache_last_push:
                last_push_time = self.cache_last_push[key]
                elapsed = time.time() - last_push_time
                if elapsed<self.min_cache_flush_delay:
                    return      
            self.push_data(key)

    def push_data(self, key):
        self.cache_last_push[key] = time.time()
        if self.aggregate_metrics=="none":
            self.push_method(key, self.cache[key])
        elif self.aggregate_metrics=="mean":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.mean(s)))])
        elif self.aggregate_metrics=="max":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.max(s)))])
        elif self.aggregate_metrics=="sum":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.sum(s)))])
        self.cache[key] = []


    def __delete__(self):
        self.stop_flag.set()
        self.timer.cancel()
        self.flush_cache()

    def flush_cache(self):
        for key in self.cache:
            self.push_data(key)



import logging

class LogHandler(logging.StreamHandler):
    """
    A handler class which allows the cursor to stay on
    one line for selected messages
    """
    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def emit(self, record):
        try:
            msg = self.format(record)
            stream = self.stream
            _msg = msg
            if len(msg)<2 or msg[-1:]!="\r":
                _msg += self.terminator
            stream.write(_msg)
            self.callback(_msg)
            # Write logs on the server right here.
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)



from .run import Run
from .config import Config
from .auth import AuthProvider

__version__ = '1.0.3'

def login():
	auth = AuthProvider()
	if auth.is_logged():
		print('User is already logged in.')
		return
	auth.login()


import requests
import os
import time
import pyhectiqlab.ops as ops
from pyhectiqlab.auth import AuthProvider

class SharedArtifactsManager():
    def __init__(self):
        self.auth_provider = AuthProvider()


    def is_logged(self):
        return self.auth_provider.is_logged()

    def refresh_token(self):
        self.auth_provider._refresh_token()
        return

    def list_projects(self):
        if self.auth_provider.is_logged()==False:
            print("User is not authentificated.")
            return
        projects = ops.get_all_projects(token=self.auth_provider.token)
        if len(projects["result"])==0:
            print("No projects.")
        else:
            for project in projects["result"]:
                print(f"[{project['id']}] {project['name']}")

    def list_artifacts(self, project_id: str):
        if self.auth_provider.is_logged()==False:
            print("User is not authentificated.")
            return
        return ops.list_shared_artifacts(project_id, token=self.auth_provider.token)

    def select_artifact(self, project_id: str):
        if self.auth_provider.is_logged()==False:
            print("User is not authentificated.")
            return
        print("Listing artifacts...")
        artifacts = ops.list_shared_artifacts(project_id, token=self.auth_provider.token)
        print("Artifacts fetched.\n")
        if len(artifacts["results"])==0:
            print("No artifacts.")
            return

        print("Select an artifact: ")
        for i,artifact in enumerate(artifacts["results"]):
            # print(i, artifact)
            print(f" [{i}] {artifact['name']}")

        index = input("Index: ")
        try:
            if int(index)>=len(artifacts["results"]):
                print("Invalid input.")
                return
        except:
            print("Invalid input.")
            return

        artifact = artifacts["results"][int(index)]
        download_path = input("Enter download path [default ./]: ")
        if download_path=="":
            download_path = "./"

        print("Downloading ...")

        res = ops.get_shared_artifacts_download_link(artifact["uuid"], token=self.auth_provider.token)
        url = res["url"]

        res = requests.get(url)
        savepath = os.path.join(os.path.abspath(download_path), artifact["name"])
        open(savepath, "wb").write(res.content)
        print(f"Saved at {savepath}")
        return

    def push_artifact(self, filepath: str, project_id:str):
        if self.auth_provider.is_logged()==False:
            print("User is not authentificated.")
            return
        filename = os.path.basename(filepath)
        num_bytes = os.path.getsize(filepath)
        content_bytes = open(filepath, "rb")

        return ops.add_shared_artifact(project_id=project_id, 
                                        filename=filename, 
                                        content_bytes=content_bytes,
                                         num_bytes=num_bytes, 
                                         token=self.auth_provider.token)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
__version__ = '1.0.3'

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

requirements = [
    "requests",
    "tqdm",
    "ipykernel",
    "traitlets",
    "jupyter_core",
    "numpy",
    "packaging",
    "click",
    "GitPython",
    "pyjwt==1.7.1"]

setup(
    name="pyhectiqlab",
    version=__version__,
    description='Python client to use the hectiq lab',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Edward Laurence',
    author_email='edwardl@hectiq.ai',
    url='https://www.hectiq.ai',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='pip requirements imports',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Science/Research',
    ],
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'hectiqlab=pyhectiqlab.cli:main',
        ],
    },
)


import threading

class RepeatedTimer(threading.Thread):
    def __init__(self, event, method, delay):
        super().__init__()
        self.daemon = True
        self.stopped = event
        self.delay = delay
        self.method = method

    def run(self):
        while not self.stopped.wait(self.delay):
            self.method()


import click
import sys
import getpass
import pyhectiqlab
from pyhectiqlab.auth import AuthProvider
from pyhectiqlab.artifacts import SharedArtifactsManager
from pyhectiqlab.mlmodels import download_mlmodel as ops_download_mlmodel
from pyhectiqlab.datasets import download_dataset as ops_download_dataset

def get_creds():
	username = input("Username: ")
	password = getpass.getpass(prompt='Password: ', stream=None) 
	return username, password

@click.group()
def cli():
    """Just a group."""
    pass

@cli.command()
def login():
	auth = AuthProvider()
	if auth.is_logged():
		click.echo("User is already logged in.")
		return

	username, password = get_creds()
	click.echo("Connecting...")
	success = auth.login_with_password(username, password)
	if success:
		click.echo("User is authentificated.")
	else:
		click.echo('Unsuccessful login.')

@cli.command()
def logout():
	auth = AuthProvider()
	auth.logout()
	click.echo('Logout completed.')

@cli.command()
def version():
	click.echo(pyhectiqlab.__version__)
	
@cli.command()
def projects():	
	manager = SharedArtifactsManager()
	manager.list_projects()

@cli.command()
@click.option('-p', '--project_id', help='id of the project', required=True)
def artifacts(project_id):
	manager = SharedArtifactsManager()
	manager.select_artifact(project_id)

@cli.command()
@click.option('-p', '--project_id', help='id of the project', required=True)
@click.option('-f', '--filename', help='Name of the shared artifact', required=True)
def post_artifact(project_id, filename):
	manager = SharedArtifactsManager()
	click.echo('Pushing artifacts.')
	manager.push_artifact(filename, project_id)
	click.echo('Artifact pushed')

@cli.command()
@click.option('-p', '--project_id', help='id of the project', required=False)
@click.option('-n', '--name', prompt='MLModel name', help='Name of the mlmodel', required=True)
@click.option('-v', '--version', help='Version of the mlmodel. If not specified, will download the latest release', required=False)
@click.option('-s', '--save_path', prompt='Save path', help='Save path', required=True)
@click.option('-o', '--overwrite', is_flag=True)
def download_mlmodel(project_id, name, version, save_path, overwrite):
	dir_path = ops_download_mlmodel(mlmodel_name=name, 
						project_id=project_id, 
						version=version, 
						save_path=save_path, 
						overwrite=overwrite)
	if dir_path is not None:
		click.echo(f'MLModel saved in {dir_path}')

@cli.command()
@click.option('-p', '--project_id', help='id of the project', required=False)
@click.option('-n', '--name', prompt='Dataset name', help='Name of the dataset', required=True)
@click.option('-v', '--version', help='Version of the dataset. If not specified, will download the latest release', required=False)
@click.option('-s', '--save_path', prompt='Save path', help='Save path', required=True)
@click.option('-o', '--overwrite', is_flag=True)
def download_dataset(project_id, name, version, save_path, overwrite):
	dir_path = ops_download_dataset(dataset_name=name, 
						project_id=project_id, 
						version=version, 
						save_path=save_path, 
						overwrite=overwrite)
	if dir_path is not None:
		click.echo(f'MLModel saved in {dir_path}')
